import { Injectable } from "@nestjs/common";
import { InjectRepository } from "@nestjs/typeorm";
import { fetchAllProductsForScheduler, fetchAllRetailers, fetchProductsCount } from "src/helper/dutchie-plus-query";
import { getGraphQLClient } from "src/helper/graphql-client-helper";
import {
    ProductCategoryRepository
} from "../dutchie-plus/custom-services/product-category/product-category.repository";
import { ProductServiceRepository } from "../dutchie-plus/custom-services/product-service/product-service.repository";
import { urlSlug } from "../../helper/common-helper";
import { Cron, CronExpression } from "@nestjs/schedule";
import {
    RetailerServiceRepository
} from "../dutchie-plus/custom-services/retailer-service/retailer-service.repository";
import { ElasticsearchService } from "@nestjs/elasticsearch";
import { ProductEffectsRepository } from "../dutchie-plus/custom-services/product-effects/product-effects.repository";

@Injectable()
export class SchedulerService {
    constructor(
      @InjectRepository(ProductCategoryRepository)
      private readonly productCatRepo: ProductCategoryRepository,
      @InjectRepository(ProductServiceRepository)
      private readonly productServiceRepo: ProductServiceRepository,
      @InjectRepository(RetailerServiceRepository)
      private readonly retailerServiceRepo: RetailerServiceRepository,
      private readonly esService: ElasticsearchService,
      @InjectRepository(ProductEffectsRepository)
      private readonly productEffectRepo: ProductEffectsRepository
    ) {
    }

    async processProductData(productData: any, retailerId: string, retailerName: string): Promise<any | Error> {
        try {
            const [
                productService,
                productCategory,
                productEffects
            ] = await Promise.all([
                this.productServiceRepo.findByFilterQuery({}),
                this.productCatRepo.findAll(),
                this.productEffectRepo.findByFilterQuery({})
            ]);

            const productServiceDatasetIds = new Set(productService.map((item) => item.productId));
            const productMapService: any = new Map(productService.map(item => [item.productId, item]));
            const productDifferences = [];
            const differeneceDeleteIds = [];
            const categoriesArr = [];
            const effectsArr = [];

            await Promise.all(productData.menu.products.map(async (item) => {
                // product service
                if (productServiceDatasetIds.size > 0) {
                    if (!productServiceDatasetIds.has(item.id)) {
                        productDifferences.push(item);
                    } else {
                        // Now you can find a product by id in constant time:
                        const matchingItem = productMapService.get(item.id);
                        // const matchingItem = productService.find((item2) => item2.productId === item.id);
                        const brand = item.brand !== null ? item.brand.name : "";
                        if (
                          matchingItem &&
                          (matchingItem.brandName !== brand ||
                            matchingItem.image !== item.image ||
                            matchingItem.productName !== item.name ||
                            matchingItem.productSlug !== item.slug ||
                            matchingItem.category !== item.category ||
                            matchingItem.subCategory !== item.subcategory ||
                            matchingItem.strainType !== item.strainType)
                        ) {
                            differeneceDeleteIds.push(item.id);
                            productDifferences.push(item);
                        }
                    }
                } else {
                    productDifferences.push(item);
                }

                // all categories
                categoriesArr.push(item.category);
                // all effects
                effectsArr.push(...item.effects);
            }));


            // if any update occers delete
            if (differeneceDeleteIds.length > 0) {
                await this.productServiceRepo.deleteRowsFromDatabase(differeneceDeleteIds);
            }

            // create entities
            console.log(retailerName, ">", " New Product = ", productDifferences.length);
            const createProductServiceDtos = productDifferences.map((item) => {
                const brand = item.brand !== null ? item.brand.name : "";
                return {
                    retailerId,
                    retailerName,
                    productId: item.id,
                    productName: item.name,
                    productSlug: item.slug,
                    image: item.image,
                    brandName: brand,
                    category: item.category,
                    subCategory: item.subcategory,
                    strainType: item.strainType,
                    variants: item.variants.length
                };
            });

            if (createProductServiceDtos.length > 0) {
                await this.productServiceRepo.createEntities(createProductServiceDtos);
            }

            // check effects
            const effects = Array.from(new Set(effectsArr));

            if (Object.keys(productEffects).length > 0) {
                const difference = effects.filter((effect) =>
                  !productEffects.some((e) => e.effects === effect)
                );
                console.log("Product Effects Difference = ", difference);
                // if difference found then create entities with difference product effects
                const createProductEffectsDtos = difference.map((item) => {
                    return {
                        effects: item
                    };
                });

                if (createProductEffectsDtos.length > 0) {
                    await this.productEffectRepo.createEntities(createProductEffectsDtos);
                }
            } else {
                // if product effects table is empty in our local database then create all the effect in local
                const createProductEffectsDtos = effects.map((item) => {
                    return {
                        effects: item
                    };
                });

                if (createProductEffectsDtos.length > 0) {
                    await this.productEffectRepo.createEntities(createProductEffectsDtos);
                }
            }

            // check category
            const categories = Array.from(new Set(categoriesArr));

            if (Object.keys(productCategory).length > 0) {
                const difference = categories.filter((category) =>
                  !productCategory.some((c) => c.categoryName === category)
                );
                console.log("Product Categories Difference = ", difference);
                // if difference found then create entities with difference product category
                const createProductCategoryDtos = difference.map((item) => {
                    return {
                        categoryName: item,
                        urlSlug: urlSlug(item)
                    };
                });

                if (createProductCategoryDtos.length > 0) {
                    await this.productCatRepo.createEntities(createProductCategoryDtos);
                }
            } else {
                // if product categories table is empty in our local database then create all the category in local
                const createProductCategoryDtos = categories.map((item) => {
                    return {
                        categoryName: item,
                        urlSlug: urlSlug(item)
                    };
                });

                if (createProductCategoryDtos.length > 0) {
                    await this.productCatRepo.createEntities(createProductCategoryDtos);
                }
            }

        } catch (e) {
            console.log(e);
            return e;
        }
    }

    async checkRetailerDifference(data: any): Promise<any | Error> {
        try {
            const retailerService = await this.retailerServiceRepo.findByFilterQuery({});

            if (Object.keys(retailerService).length > 0) {
                const existingRetailerIds = retailerService.map(x => x.retailerId);
                const difference = data.retailers.filter(obj => !existingRetailerIds.includes(obj.id));
                console.log("Retailer Difference =", difference);

                const dtoObjects = difference.map(obj => ({
                    retailerId: obj.id,
                    retailerName: obj.name,
                    retailerAddress: JSON.stringify(obj.addressObject)
                }));

                await this.retailerServiceRepo.createEntities(dtoObjects);
            } else {
                const dtoObjects = data.retailers.map(obj => ({
                    retailerId: obj.id,
                    retailerName: obj.name,
                    retailerAddress: JSON.stringify(obj.addressObject)
                }));

                await this.retailerServiceRepo.createEntities(dtoObjects);
            }

        } catch (e) {
            console.log(e);
            return e;
        }
    }

    async checkRetailerDifferenceArizona(data: any): Promise<any | Error> {
        try {
            const retailerService = await this.retailerServiceRepo.findByFilterQuery({});

            if (Object.keys(retailerService).length > 0) {
                const existingRetailerIds = retailerService.map(x => x.retailerId);
                const difference = data.retailers.filter(obj => !existingRetailerIds.includes(obj.id));
                console.log("Retailer Difference =", difference);

                const dtoObjects = difference.map(obj => ({
                    retailerId: obj.id,
                    retailerName: obj.name,
                    retailerAddress: JSON.stringify(obj.addressObject)
                }));

                await this.retailerServiceRepo.createEntities(dtoObjects);
            } else {
                const dtoObjects = data.retailers.map(obj => ({
                    retailerId: obj.id,
                    retailerName: obj.name,
                    retailerAddress: JSON.stringify(obj.addressObject)
                }));

                await this.retailerServiceRepo.createEntities(dtoObjects);
            }

        } catch (e) {
            console.log(e);
            return e;
        }
    }


    // @Cron(CronExpression.EVERY_3_HOURS)
    async createUpdateDutchiePlusProductsAndCategory() {
        try {
            console.log(">>>Run Create or Update Dutchie Plus Products Scheduler<<<");

            // Get GraphQL client
            const graphQLClient = getGraphQLClient();

            // Fetch all retailers
            const query = fetchAllRetailers();
            const data = await graphQLClient.request(query);
            const retailers = data.retailers;

            // Check if retailers exist
            if (retailers.length > 0) {
                // Create an array of promises for fetching product data
                console.log("Start Fetching product count Parallel  =", new Date());
                const fetchProductsCountObject = fetchProductsCount();
                const retailerListPromises = retailers.map((retailer) =>
                  graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
                    .then(response => {
                        if (response && response.menu && response.menu.productsCount != null) {

                            return {
                                retailerId: retailer.id,
                                retailerName: retailer.name,
                                pagination: {
                                    offset: 0,
                                    limit: response.menu.productsCount ? parseInt(response.menu.productsCount) : 0
                                }
                            };
                        }
                    })
                );

                let allRetailerProductCount = await Promise.all(retailerListPromises);
                allRetailerProductCount = allRetailerProductCount.filter((data) => data.pagination.limit > 0);
                console.log("End Fetching product count Parallel =", new Date());


                console.log("Start Fetching product Parallel End =", new Date());
                const fetchProductsQuery = fetchAllProductsForScheduler();
                const fetchProductPromises = allRetailerProductCount.map(variable =>
                  graphQLClient.request(fetchProductsQuery, variable).then(response => {
                      if (response && response.menu && response.menu.products) {
                          return {
                              retailerId: variable.retailerId,
                              retailerName: variable.retailerName,
                              productData: response
                          };
                      }
                  })
                );
                console.log("End Fetching product Parallel End =", new Date());

                const allProductData = await Promise.all(fetchProductPromises);
                //console.log(allProductData)
                for (const data of allProductData) {
                    console.log("Send for Processing", new Date());
                    await this.processProductData(data.productData, data.retailerId, data.retailerName);
                    console.log("End processing =", new Date());
                }
                console.log(">>>End Create or Update Dutchie Plus Products Scheduler<<<");
            } else {
                console.log("Retailers Not Found!");
            }
        } catch (e) {
            console.log(e);
            return e;
        }
    }


    // @Cron(CronExpression.EVERY_5_HOURS)
    async createUpdateRetailer() {
        try {
            const graphQLClient = getGraphQLClient();
            const query = fetchAllRetailers();
            const data = await graphQLClient.request(query);

            // store retailers in local storage
            await this.checkRetailerDifference(data);
        } catch (e) {
            console.log(e);
            return e;
        }
    }

    private async indexIsExists(index: string) {
        return await this.esService.indices.exists({ index });
    }

    private async deleteIndex(index: string) {
        const isExists = await this.indexIsExists(index);
        if (isExists) {
            await this.esService.indices.delete({ index });
        }
    }

    private async createIndex(index: string, body: any) {
        const isExists = await this.indexIsExists(index);
        if (!isExists) {
            const data = await this.esService.indices.create({ index, body });
            return data.acknowledged;
        }
        return false;
    }

    private async fetchAllProductsForElasticSearch() {
        try {
            console.log(">>>Run Fetch All Products For Elastic Search<<<");

            // Get GraphQL client
            const graphQLClient = getGraphQLClient();

            // Fetch all retailers
            const query = fetchAllRetailers();
            const data = await graphQLClient.request(query);
            const retailers = data.retailers;

            // Check if retailers exist
            if (retailers.length > 0) {
                // Create an array of promises for fetching product data
                console.log("Start Fetching product count Parallel  =", new Date());
                const fetchProductsCountObject = fetchProductsCount();
                const retailerListPromises = retailers.map((retailer) =>
                  graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
                    .then(response => {
                        if (response && response.menu && response.menu.productsCount != null) {
                            return {
                                retailerId: retailer.id,
                                retailerName: retailer.name,
                                pagination: {
                                    offset: 0,
                                    limit: response.menu.productsCount ? parseInt(response.menu.productsCount) : 0
                                }
                            };
                        }
                    })
                );

                let allRetailerProductCount = await Promise.all(retailerListPromises);
                allRetailerProductCount = allRetailerProductCount.filter((data) => data.pagination.limit > 0);
                console.log("End Fetching product count Parallel =", new Date());

                console.log("Start Fetching product Parallel =", new Date());
                const fetchProductsQuery = fetchAllProductsForScheduler();
                const fetchProductPromises = allRetailerProductCount.map(variable =>
                  graphQLClient.request(fetchProductsQuery, variable).then(response => {
                      if (response && response.menu && response.menu.products) {
                          return {
                              retailerId: variable.retailerId,
                              retailerName: variable.retailerName,
                              productData: response
                          };
                      }
                  })
                );
                console.log("End Fetching product Parallel =", new Date());

                // console.table(allRetailerProductCount);

                const allProductData = await Promise.all(fetchProductPromises);
                const products = [];
                for (const data of allProductData) {
                    const product = data.productData.menu.products.map(x => {
                        const { id, brand, name, slug, variants, ...restData } = x;
                        const brandName = brand?.name ?? "";
                        const brandNameAndProductNameAndCategoryName = name.trim().startsWith(brandName.trim())
                          ? `${name} ${x.category}` : `${brandName} ${name} ${x.category}`;

                        return {
                            ...restData,
                            productId: id,
                            productName: name,
                            productSlug: slug,
                            brandName,
                            variants: variants?.length ?? 0,
                            retailerId: data.retailerId,
                            retailerName: data.retailerName,
                            brandNameAndProductNameAndCategoryName
                        };
                    });
                    products.push(...product);
                }
                console.log(">>>End Fetch All Products For Elastic Search<<<");
                // console.table(products.length);
                return products;
            } else {
                console.log("Retailers Not Found!");
            }
        } catch (e) {
            console.log(e);
            return e;
        }
    }

    // @Cron(CronExpression.EVERY_HOUR)
    public async populateElasticSearchProduct() {
        let indexValue: string;
        if (process.env.ENVIRONMENT == "local") {
            indexValue = process.env.PRODUCT_INDEX_DEV;
        } else {
            indexValue = process.env.PRODUCT_INDEX_PROD;
        }
        const products: any = await this.fetchAllProductsForElasticSearch();

        await this.deleteIndex(indexValue);

        const body = {
            settings: {
                analysis: {
                    filter: {
                        autocomplete_filter: {
                            type: "edge_ngram",
                            min_gram: 1,
                            max_gram: 15
                        }
                    },
                    analyzer: {
                        autocomplete: {
                            type: "custom",
                            tokenizer: "standard",
                            filter: ["lowercase", "autocomplete_filter"]
                        }
                    }
                }
            },

            mappings: {
                properties: {
                    category: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    subcategory: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    strainType: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    image: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 512
                            }
                        }
                    },
                    productId: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    productName: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    brandName: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    brandNameAndProductNameAndCategoryName: {
                        type: "text",
                        analyzer: "autocomplete",
                        search_analyzer: "standard",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 512
                            }
                        }
                    },
                    productSlug: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    retailerId: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    retailerName: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    variants: {
                        type: "long"
                    }
                }
            }
        };
        await this.createIndex(indexValue, body);
        console.log(">>>Start Bulk Insert<<<");
        const operations = products.flatMap((doc) => [
            { index: { _index: indexValue } },
            doc
        ]);

        await this.esService.bulk({ refresh: true, operations });
        console.log(">>>End Bulk Insert<<<");
    }

    private async getRetailerData() {
        const graphQLClient = getGraphQLClient();
        const query = fetchAllRetailers();
        const data = await graphQLClient.request(query);
        return data.retailers;
    }

    // @Cron(CronExpression.EVERY_6_HOURS)
    public async populateElasticSearchRetailer() {
        console.log(">>>Start Bulk Insert Retailer Service<<<", new Date());
        let indexValue: string;
        if (process.env.ENVIRONMENT == "local") {
            indexValue = process.env.RETAILER_INDEX_DEV;
        } else {
            indexValue = process.env.RETAILER_INDEX_PROD;
        }
        const retailers = await this.getRetailerData();
        // console.log(retailers);

        await this.deleteIndex(indexValue);

        const body = {
            settings: {
                analysis: {
                    filter: {
                        autocomplete_filter: {
                            type: "edge_ngram",
                            min_gram: 1,
                            max_gram: 15
                        }
                    },
                    analyzer: {
                        autocomplete: {
                            type: "custom",
                            tokenizer: "standard",
                            filter: ["lowercase", "autocomplete_filter"]
                        }
                    }
                }
            },

            mappings: {
                properties: {
                    retailerId: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    state: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    retailerName: {
                        type: "text",
                        analyzer: "autocomplete",
                        search_analyzer: "standard",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    searchKey: {
                        type: "text",
                        analyzer: "autocomplete",
                        search_analyzer: "standard",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 512
                            }
                        }
                    }
                }
            }
        };

        await this.createIndex(indexValue, body);

        const operations = retailers.flatMap((doc) => [
            { index: { _index: indexValue } },
            {
                retailerId: doc.id,
                retailerName: doc.name,
                state: doc.addressObject.state,
                searchKey: `${doc.name} ${doc.addressObject.postalCode}`
            }
        ]);
        await this.esService.bulk({ refresh: true, operations });
        console.log(">>>End Bulk Insert RetailerService<<<", new Date());
    }
}
